/* Copyright (C) 2004 Stephane GALLAND <stephane.galland@utbm.fr>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
                                                                               
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   This program is free software; you can redistribute it and/or modify
*/

#ifndef __INTERPRETER_C__
#error Can not compile this file outside interpreter.c
#endif

/* Format and replies a declaration component for
 * the assembler declaration.
 *
 * @param __d__ is the data to format
 */
#define PASM_SDECL(__d__) ((long)(__d__))

/** Format and replies a declaration of a assembler function.
 *
 * @param __func__ is the name of the assembler instruction.
 * @param __proto__ is the prototyping flags.
 */
#define PASM_DECL(__func__,__proto__) \
  PASM_SDECL(ASM_##__func__), PASM_SDECL(__proto__), PASM_SDECL(#__func__), PASM_SDECL(F_##__func__)

/** Starting code for a assembler function.
 */
#define _PASM_DEFINITION_DEFAULT_CODE(__func__)				\
  const char __CURRENTASM__[] = #__func__ ;				\
  unsigned char __PROTOTYPE__ = ASM_getInstructionPrototype(ASM_##__func__) ; \
  VIRTUAL_MEMORY* mem = VM_getVirtualMemory(vm) ;			\
  if (mem==NULL) {							\
    ASMERROR(ERR_ASM_NOTMEMMANAGER);					\
    return FALSE ;							\
  }

/** Translate one memory address.
 */
#define _PASM_DEFINITION_TRANSLATE_ADR(__adr__) if (__PROTOTYPE__&PASM_ADR1) { \
    boolean __addressredirection = FALSE ;				\
    if ( PASM_ISADRREDIRECTION(__adr__) ) {				\
      (__adr__) = PASM_UNMAKEADRREDIRECTION(__adr__) ;			\
      __addressredirection = TRUE ;					\
    }									\
    if ( PASM_ISREGISTER(__adr__) ) {					\
      MEMORY_ADDRESS baseadr,endadr ;					\
      ASSEMBLER_PARAM reg = PASM_UNMAKEREGISTER(__adr__) ;		\
      if (!VM_getArrayAddresses(vm,&baseadr,&endadr)) {			\
	ASMERROR(ERR_ASM_NODECLARE);					\
	return FALSE ;							\
      }									\
      else if (((baseadr+reg)<baseadr)||				\
	       ((baseadr+reg)>endadr)) {				\
	ASMERROR(ERR_ASM_INVALIDREGISTER);				\
	return FALSE ;							\
      }									\
      else {								\
	(__adr__) = baseadr+reg;					\
      }									\
    }									\
    else if ( PASM_ISABSOLUTEADDRESS(__adr__) ) {			\
      (__adr__) = PASM_UNMAKEABSOLUTEADDRESS(__adr__) ;			\
    }									\
    if (__addressredirection) {						\
      if (!MEM_isValidAdr(mem,(__adr__))) {				\
	ASMERROR(ERR_ASM_INVALIDADDRESS) ;				\
	return FALSE ;							\
      }									\
      (__adr__) = MEM_get(mem,__adr__) ;				\
    }									\
    if (!MEM_isValidAdr(mem,(__adr__))) {				\
      ASMERROR(ERR_ASM_INVALIDADDRESS) ;				\
      return FALSE ;							\
    }									\
  }

/** Translate two memory addresses.
 */
#define _PASM_DEFINITION_TRANSLATE_ADRS(__adr1__,__adr2__) \
  if (__PROTOTYPE__&(PASM_ADR1|PASM_ADR2)) {				\
    MEMORY_ADDRESS baseadr,endadr ;					\
    boolean __addressredirection = FALSE ;				\
    if (!VM_getArrayAddresses(vm,&baseadr,&endadr)) {			\
      ASMERROR(ERR_ASM_NODECLARE);					\
      return FALSE ;							\
    }									\
    if ( PASM_ISADRREDIRECTION(__adr1__) ) {				\
      (__adr1__) = PASM_UNMAKEADRREDIRECTION(__adr1__) ;		\
      __addressredirection = TRUE ;					\
    }									\
    if ( PASM_ISREGISTER(__adr1__) ) {					\
      ASSEMBLER_PARAM reg = PASM_UNMAKEREGISTER(__adr1__) ;		\
      if (((baseadr+reg)<baseadr)||					\
	  ((baseadr+reg)>endadr)) {					\
	ASMERROR(ERR_ASM_INVALIDREGISTER);				\
	return FALSE ;							\
      }									\
      else {								\
	(__adr1__) = baseadr+reg;					\
      }									\
    }									\
    else if ( PASM_ISABSOLUTEADDRESS(__adr1__) ) {			\
      (__adr1__) = PASM_UNMAKEABSOLUTEADDRESS(__adr1__) ;		\
    }									\
    if (__addressredirection) {						\
      if (!MEM_isValidAdr(mem,(__adr1__))) {				\
	ASMERROR(ERR_ASM_INVALIDADDRESS) ;				\
	return FALSE ;							\
      }									\
      (__adr1__) = MEM_get(mem,__adr1__) ;				\
    }									\
    if (!MEM_isValidAdr(mem,(__adr1__))) {				\
      ASMERROR(ERR_ASM_INVALIDADDRESS) ;				\
      return FALSE ;							\
    }									\
    __addressredirection = FALSE ;					\
    if ( PASM_ISADRREDIRECTION(__adr2__) ) {				\
      (__adr2__) = PASM_UNMAKEADRREDIRECTION(__adr2__) ;		\
      __addressredirection = TRUE ;					\
    }									\
    if ( PASM_ISREGISTER(__adr2__) ) {					\
      ASSEMBLER_PARAM reg = PASM_UNMAKEREGISTER(__adr2__) ;		\
      if (((baseadr+reg)<baseadr)||					\
	  ((baseadr+reg)>endadr)) {					\
	ASMERROR(ERR_ASM_INVALIDREGISTER);				\
	return FALSE ;							\
      }									\
      else {								\
	(__adr2__) = baseadr+reg;					\
      }									\
    }									\
    else if ( PASM_ISABSOLUTEADDRESS(__adr2__) ) {			\
      (__adr2__) = PASM_UNMAKEABSOLUTEADDRESS(__adr2__) ;		\
    }									\
    if (__addressredirection) {						\
      if (!MEM_isValidAdr(mem,(__adr2__))) {				\
	ASMERROR(ERR_ASM_INVALIDADDRESS) ;				\
	return FALSE ;							\
      }									\
      (__adr2__) = MEM_get(mem,__adr2__) ;				\
    }									\
    if (!MEM_isValidAdr(mem,(__adr2__))) {				\
      ASMERROR(ERR_ASM_INVALIDADDRESS) ;				\
      return FALSE ;							\
    }									\
  }

/** Defines an assembler function for a no-parameter instruction.
 *
 * @param __func__ is the name of the assembler instruction.
 */
#ifdef __DEBUG__
#define PASM_DEF(__func__)			\
  boolean F_##__func__(VIRTUAL_MACHINE* vm) {	  \
  _PASM_DEFINITION_DEFAULT_CODE(__func__);	  \
  d_dbglog("\t\t* run: %s\n", __CURRENTASM__) ;
#else
#define PASM_DEF(__func__)			\
  boolean F_##__func__(VIRTUAL_MACHINE* vm) {	\
  _PASM_DEFINITION_DEFAULT_CODE(__func__) ;
#endif

/** Defines an assembler function for a no-parameter instruction.
 *
 * @param __func__ is the name of the assembler instruction.
 * @param __p1__ is the first parameter
 */
#ifdef __DEBUG
#define PASM_DEF1(__func__,__p1__)					\
  boolean F_##__func__(VIRTUAL_MACHINE* vm, ASSEMBLER_PARAM __p1__) {	\
    _PASM_DEFINITION_DEFAULT_CODE(__func__) ;				\
    d_dbglog("\t\t* run: %s 0x%lX\n", __CURRENTASM__, (unsigned long)(__p1__)) ; \
    _PASM_DEFINITION_TRANSLATE_ADR(__p1__) ;
#else
#define PASM_DEF1(__func__,__p1__) \
  boolean F_##__func__(VIRTUAL_MACHINE* vm, ASSEMBLER_PARAM __p1__) {	\
  _PASM_DEFINITION_DEFAULT_CODE(__func__) ;				\
  _PASM_DEFINITION_TRANSLATE_ADR(__p1__) ;
#endif

/** Defines an assembler function for a no-parameter instruction.
 *
 * @param __func__ is the name of the assembler instruction.
 * @param __p1__ is the first parameter
 */
#ifdef __DEBUG
#define PASM_DEF2(__func__,__p1__,__p2__)				\
  boolean F_##__func__(VIRTUAL_MACHINE* vm, ASSEMBLER_PARAM __p1__, ASSEMBLER_PARAM __p2__) { \
  _PASM_DEFINITION_DEFAULT_CODE(__func__) ;				\
  d_dbglog("\t\t* run: %s 0x%lX 0x%lX\n", __CURRENTASM__,		\
	   (unsigned long)(__p1__),(unsigned long)(__p2__)) ;		\
  _PASM_DEFINITION_TRANSLATE_ADRS(__p1__,__p2__) ;
#else
#define PASM_DEF2(__func__,__p1__,__p2__)				\
  boolean F_##__func__(VIRTUAL_MACHINE* vm, ASSEMBLER_PARAM __p1__, ASSEMBLER_PARAM __p2__) { \
  _PASM_DEFINITION_DEFAULT_CODE(__func__) ;				\
  _PASM_DEFINITION_TRANSLATE_ADRS(__p1__,__p2__) ;
#endif

/** Define the end of a definition.
 */
#define PASM_END __PROTOTYPE__ = 0 ; return FALSE ; }
