/* Copyright (C) 2004 Stephane GALLAND <stephane.galland@utbm.fr>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
                                                                               
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   This program is free software; you can redistribute it and/or modify
*/
#ifndef __EW_TYPES_H__
#define __EW_TYPES_H__

/** Type that representes the quantity of opponents.
 */
typedef unsigned char OPPONENT_SIZE;

/** Defines the type which representes the size of a source code.
 */
typedef unsigned short CODE_SIZE ;

/** Type of ONE element of the memory.
 */
typedef unsigned long MEMORY_ITEM ;

/** Type which representes an memory address.
 */
typedef unsigned long MEMORY_ADDRESS ;

/** Type which representes an memory size.
 */
typedef unsigned long MEMORY_SIZE ;

/** Register index flag
 */
#define PASM_REGISTERFLAG  (1<<(sizeof(ASSEMBLER_PARAM)*8-1))

/** Replies if the specified parameter is a register index
 */
#define PASM_ISREGISTER(__reg__)  (((__reg__)!=VNULL)&&((PASM_REGISTERFLAG)&(__reg__)))

/** Build a register index
 */
#define PASM_MAKEREGISTER(__reg__)  ((PASM_REGISTERFLAG)|(__reg__))

/** Extract a register index
 */
#define PASM_UNMAKEREGISTER(__reg__)  ((~PASM_REGISTERFLAG)&(__reg__))

/** Memory address flag
 */
#define PASM_ABSOLUTEADDRESSFLAG  (1<<(sizeof(ASSEMBLER_PARAM)*8-2))

/** Replies if the specified parameter is an absolute memory address
 */
#define PASM_ISABSOLUTEADDRESS(__adr__)  (((__adr__)!=VNULL)&&((PASM_ABSOLUTEADDRESSFLAG)&(__adr__)))

/** Build an absolute address
 */
#define PASM_MAKEABSOLUTEADDRESS(__adr__)  ((PASM_ABSOLUTEADDRESSFLAG)|(__adr__))

/** Extract an aboluste address
 */
#define PASM_UNMAKEABSOLUTEADDRESS(__adr__)  ((~PASM_ABSOLUTEADDRESSFLAG)&(__adr__))

/** Data segment flag
 */
#define PASM_DATASEGMENTFLAG  (1<<(sizeof(ASSEMBLER_PARAM)*8-3))

/** Replies if the specified parameter is a data segment
 */
#define PASM_ISDATASEGMENT(__dt__)  (((__dt__)!=VNULL)&&((PASM_DATASEGMENTFLAG)&(__dt__)))

/** Build a data segment
 */
#define PASM_MAKEDATASEGMENT(__dt__)  ((PASM_DATASEGMENTFLAG)|(__dt__))

/** Extract a data segment
 */
#define PASM_UNMAKEDATASEGMENT(__dt__)  ((~PASM_DATASEGMENTFLAG)&(__dt__))

/** Address redirection flag
 */
#define PASM_ADRREDIRECTIONFLAG  (1<<(sizeof(ASSEMBLER_PARAM)*8-4))

/** Replies if the specified parameter is an address redirection
 */
#define PASM_ISADRREDIRECTION(__dt__)  (((__dt__)!=VNULL)&&((PASM_ADRREDIRECTIONFLAG)&(__dt__)))

/** Build an address redirection
 */
#define PASM_MAKEADRREDIRECTION(__dt__)  ((PASM_ADRREDIRECTIONFLAG)|(__dt__))

/** Extract an address redirection
 */
#define PASM_UNMAKEADRREDIRECTION(__dt__)  ((~PASM_ADRREDIRECTIONFLAG)&(__dt__))
/** Theorical memory limit.
 */
#define THEORICAL_MEMORY_LIMIT \
  ( ~( PASM_ADRREDIRECTIONFLAG |		\
       PASM_DATASEGMENTFLAG |			\
       PASM_ABSOLUTEADDRESSFLAG |		\
       PASM_REGISTERFLAG ) )

#endif
